# Copyright (C) 2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
"""
Constants for file/disk sizes.
"""

KiB = 1024
MiB = 1024**2
GiB = 1024**3
TiB = 1024**4
PiB = 1024**5
